#ifndef __main_h
 #define __main_h

#include "accumulator.h"


enum editorAttributes {
  HL_NORMAL = 20,
  HL_MATCH,
  
  HL_DIGIT, HL_STRING, HL_COMMENT, HL_KEYWORD, HL_TYPE, HL_DIRECTIVE,
  
  HL_RB1, HL_RB2, HL_RB3, HL_RB4, HL_RB5, HL_RB6, HL_RB7, HL_RB8,
    
  END_LIST_ATTRIB
};

enum focusPoints {  FOCUS_PROMPT, FOCUS_DOCUMENT };


int BRACKETRAINBOW[] = {  HL_RB1,  HL_RB2,  HL_RB3,  HL_RB4,  HL_RB5,  HL_RB6,  HL_RB7,  HL_RB8 }; // values for rainbow cycle




typedef struct erow {
  char *chars;
  char *render;
  char *attrib;
} erow;



typedef struct DocLoc_s {
   uint32_t      cx, cy;         // cursor x,y in file
   int           rowoff,coloff;  // scroll offsets   
} DocLoc_t;



extern struct Editor_s;  // heh, bit of a circle here...

typedef struct Dialog_s {

   buffEdProc_t  el;             // prompt input processor
   char *prompt;
   char *hint;
   void (*EV_CB)       (struct Editor_s * E, int); // Editor_t is the link back to the main data, int is the event

} Dialog_t;



typedef struct Editor_s {
   int           windowrows, windowcols;
   
   char *        filename;       // filename
//   erow         *row;            // array of parsed file rows
   AOffset_t     lineCount;        // number of rows    
   
   tac_t         accum;
   
   DocLoc_t      crs_curr, crs_bkp;
   
   int           rx;             // rendered cursor x position

   uint8_t       dirty;          // content edited flag       
   
   char *        tmpMessage;     // temporary message text
   time_t        messageExpire;  // expire time of temp message
      
   Dialog_t      dialog;
   
   int           focus;
   
} Editor_t;



typedef struct substTable_s { // character to string substitution system, only used to freaking tab characters
  char from;
  char * to;
} substTable_t;




char *       editorAttribToESC    ( int a );

// modify

void editorOpenCallback     ( Editor_t * E, int c);
int  editorOpen             ( Editor_t * E, char *filename);

void editorSaveCallback     ( Editor_t * E, int c);
void editorCmdSave          ( Editor_t * E );
void editorCmdSaveAs        ( Editor_t * E) ;
int  editorSave             ( Editor_t * E, char *filename) ;

void editorFindCallback     ( Editor_t * E, int key ) ;
void editorFind             ( Editor_t * E );

void editorNew              ( Editor_t * E );
void editorCmdOpen          ( Editor_t * E );
void initEditor             ( Editor_t * E );
void die                    ( const char *s);

void promptProcessEvent     ( Editor_t * E, int c ) ;
void editorProcessEvent     ( Editor_t * E, int c );

void editorMoveCursor       ( Editor_t * E, int c);

int  promptProcessKeypress  ( Editor_t * E, buffEdProc_t * this, int c);


char * genAttribs           ( Editor_t * E, erow *row) ;
void editorRefreshScreen    ( Editor_t * E );
void editorScroll           ( Editor_t * E );

void editorInsertChar       ( Editor_t * E, int c);
void editorDelChar          ( Editor_t * E );

/* (message and status)  */
//char *editorPrompt          ( Editor_t * E, char *prompt, char *hint, char *dfValue, void (*callback)(Editor_t * E, char *, int)) ;
char *editorPrompt          ( Editor_t * E, char *prompt, char *hint, char *dfValue, void (*callback)(Editor_t * E,  int)) ;

void editorDrawStatusBar    ( Editor_t * E, char **ab);
void editorSetStatusMessage ( Editor_t * E, const char *fmt, ...);

void editorDrawMessageBar   ( Editor_t * E, char **ab);
void editorInsertFlow       ( Editor_t * E, int c);

void editorBackspaceChar( Editor_t * E );

// drop
void editorDrawRows         ( Editor_t * E, char ** ab);
void editorInsertRow        ( Editor_t * E, int at, char *s, size_t len) ;
char *editorRowsToString    ( Editor_t * E );
void editorDelRow           ( Editor_t * E, int at ) ;
void editorInsertNewline    ( Editor_t * E );
void editorUpdateRow        ( Editor_t * E, erow *row);
void editorRowInsertChar    ( Editor_t * E, erow *row, unsigned int at, int c);
void editorRowDelChar       ( Editor_t * E, erow *row, int at);
void editorFreeRow          ( Editor_t * E, erow *row ) ;
void editorRowAppendString  ( Editor_t * E, erow *row, char *s );
void editorRowInsertString  ( Editor_t * E, erow *row, unsigned int at, char *s);



#endif


