// ANSI terminal io.

#include "terminal.h"

Status_t enableRawMode() {
  struct termios raw;

  if (tcgetattr(STDIN_FILENO, &C.orig_termios) == -1) return ExecFail;
  atexit((void *)disableRawMode);
  
  raw = C.orig_termios;
  raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  raw.c_oflag &= ~(OPOST);
  raw.c_cflag |= (CS8);
  raw.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
  raw.c_cc[VMIN]  = 0;
  raw.c_cc[VTIME] = 1;
  
  if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw) == -1) return ExecFail;
  
  writes(ANSI_MouseOn);
  
}


Status_t disableRawMode() {
   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &C.orig_termios) == -1)
    return ExecFail;
    
   writes(ANSI_MouseOff); 
   writes(ANSI_ColorRST);
}



int getWindowSize(int *rows, int *cols) {
  struct winsize ws;
  
  if ((ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) == -1) || (ws.ws_col == 0)) {
    return -1;
  } else {
    *cols = ws.ws_col;
    *rows = ws.ws_row;
    return 0;
  }
}


/// ------------------------------------------------------------------------------------------------


void writes( char * s) {
  write(STDOUT_FILENO, s, strlen(s));
}


///-------------------------------------------------------------------------------------------------


// part of an event loop
int checkStdin(void) { 

  int    rv;
  int fsmo;
  char   cp[10];
  fd_set rfds;
  struct timeval tv;
  int n;

  tv.tv_sec  = 0 ;  tv.tv_usec = 250000;

 // while(1){

      FD_ZERO(&rfds);
      FD_SET(STDIN_FILENO, &rfds); 

      rv = select(STDIN_FILENO+1, &rfds, NULL, NULL, &tv); 
      if (0) {
      } else if (rv == -1) {
         perror("select()");
      } else if (FD_ISSET(STDIN_FILENO, &rfds)){              
         rv = read(STDIN_FILENO, &cp, 10);                     // get  character
         if ((rv == -1) && (errno != EAGAIN)) return KEY_NONE; // die("read"); !?!?!?!
         
         if (cp[0] == '\e') {// was it escape???  
             if (rv == 1) {
               return '\e';  // just a normal escape, dont let the state machine see that!                                
             } else {
               for ( n = 0; n < rv; n++) 
                 if (( fsmo = stdinFSM( cp[n] )) != KEY_NONE) return fsmo;
               return KEY_NONE;               
             }          
         } else {
           return stdinFSM( cp[0] ); // feed it into the state machine
         }
      } 
      
      return KEY_NONE; // cycle for updates.
 
 // }

}



/*

<char>                                -> char
<esc>                                 -> esc
<esc> <esc>                           -> esc
<esc> <char>                          -> Alt-keypress or keycode sequence
<esc> '['                             -> Alt-[
<esc> '[' <num>
<esc> '[' (<num>) (';'<num>) '~'      -> keycode sequence, <num> defaults to 1
<esc> 'O' <char>

*/



int stdinFSM(char  c) { // parse characters in, escape code or otherwise
  
  static int state = 0;
  static char b[32];
  static int i = 0;
  
  switch (state) {
  
    case 0: // first character
      if ((c >= 0x20) && (c <= 0x7F))    return c;     // normal character
      if  (c != '\e')                    return extCharLookup(c); 
      // problem here, single escape? or escape sequence?
      state = 1;
    break;
    
    case 1:
      if (c == '\e')  {                   state = 0;   return '\e'; } // double escape
      if ((c == '[' ) || (c == 'O' ) )  { state = 2;   b[0] = c;  i = 1;  }
      else state = 0;   // error
    break;    

    case 2:
      if (0) {
      } else if (c == 'M')  {                   state = 10;          i = 0;    // mouse event  
      } else if( (c == ';') || (isdigit(c)) ) { state = 3; b[i] = c; i++;  
      } else {
        b[i] = c; i++; b[i] = 0;
        state = 0;
        return escKeyLookup( b ); 
      }
    break;
    
    case 3:
      if (0) {
      } else if( (c == ';') || (isdigit(c)) ) { 
        b[i] = c; i++;  
      } else {
        b[i] = c; i++; b[i] = 0;        
        state = 0;
        return escKeyLookup( b ); 
      }      
    break;
                    
    case 10:
    case 11:
      b[i] = c;
      i++;
      state++;
    break;    
  
    case 12:
      b[i] = c;
      state = 0;
      M.buttons = b[0];  M.X = b[1]-33; M.Y = b[2]-33;
      return MOUSE_EVENT;
    break;
  
  }

  return KEY_NONE;

}



int escKeyLookup (char *s) {

 // fprintf(stderr, "%s\n", (s[0]=='\e')?&s[1]:&s[0]);

  if (0) {
  } else if (!strcmp( s, "[A"))    {   return KEY_UP;       
  } else if (!strcmp( s, "[B"))    {   return KEY_DOWN;
  } else if (!strcmp( s, "[C"))    {   return KEY_RIGHT;
  } else if (!strcmp( s, "[D"))    {   return KEY_LEFT;
  } else if (!strcmp( s, "[H"))    {   return KEY_HOME;
  } else if (!strcmp( s, "[F"))    {   return KEY_END;
  } else if (!strcmp( s, "[3~"))   {   return KEY_DEL; // \e[3~
  } else if (!strcmp( s, "[1~"))   {   return KEY_HOME;
  } else if (!strcmp( s, "[7~"))   {   return KEY_HOME;
  } else if (!strcmp( s, "[4~"))   {   return KEY_END;
  } else if (!strcmp( s, "[8~"))   {   return KEY_END;          
  } else if (!strcmp( s, "[5~"))   {   return PAGE_UP;
  } else if (!strcmp( s, "[6~"))   {   return PAGE_DOWN;
  } else if (!strcmp( s, "OR"))    {   return KEY_F3;
  } else if (!strcmp( s, "[13~"))  {   return KEY_F3;
  } else if (!strcmp( s, "[[C"))   {   return KEY_F3;       // terminal // oh cmon now!
  } else if (!strcmp( s, "[25~"))  {   return KEY_SHIFT_F3;
  } else if (!strcmp( s, "[1;2R")) {   return KEY_SHIFT_F3;
  } else if (!strcmp( s, "[37~"))  {   return KEY_SHIFT_F3; // terminal
  } else if (!strcmp( s, "[1;2H")) {   return SHIFT_HOME;
  } else if (!strcmp( s, "[1;2F")) {   return SHIFT_END;
  } else if (!strcmp( s, "[1;2A")) {   return SHIFT_UP;
  } else if (!strcmp( s, "[1;2B")) {   return SHIFT_DOWN;
  } else if (!strcmp( s, "[1;2C")) {   return SHIFT_RIGHT;
  } else if (!strcmp( s, "[1;2D")) {   return SHIFT_LEFT;      
  
  } else {
    fprintf(stderr, "match fail %s\n", s ); // show what we couldn't match
    return 0;
  }
 
}



unsigned int extCharLookup (unsigned char c) {

  switch(c) {
    case CTRL_KEY('a'): return CTRL_A;
    case CTRL_KEY('f'): return CTRL_F;
    case CTRL_KEY('q'): return CTRL_Q;
    case CTRL_KEY('l'): return CTRL_L; // I have no idea.
    case CTRL_KEY('s'): return CTRL_S;
    case CTRL_KEY('n'): return CTRL_N;
    case CTRL_KEY('o'): return CTRL_O;    
    case 0x0D:          return c;    
    case 0x0A:          return c;
    case 0x08:          return c;     //backspace
    case 0x09:          return c;     // tab!
    case 0xB0:          return ALT_0;
    case 0xB1:          return ALT_1;
    case 0xB2:          return ALT_2;
    case 0xB3:          return ALT_3;
    case 0xB4:          return ALT_4;
  
    default:           
      fprintf(stderr, "WTF key: %X\n", c);
      return 0;
  }

}



